package model

import javax.swing.plaf.ColorUIResource

const val COLOR_RESET = "\u001B[0m"
const val COLOR_BLACK = "\u001B[30m"
const val COLOR_RED = "\u001B[31m"
const val COLOR_GREEN = "\u001B[32m"
const val COLOR_YELLOW = "\u001B[33m"
const val COLOR_BLUE = "\u001B[34m"
const val COLOR_PURPLE = "\u001B[35m"
const val COLOR_CYAN = "\u001B[36m"
const val COLOR_WHITE = "\u001B[37m"

const val heartSymbol = "♥"

// TODO: Heritage
class HeartCard(isFaceUp: Boolean, cardValue: CardValue) : Card(isFaceUp, cardValue) {

    constructor(isFaceUp: Boolean, value: Int) : this(isFaceUp, CardValue.valueOf(value))
    constructor(isFaceUp: Boolean, value: String): this(isFaceUp, CardValue.fromString(value))

    override val colorConsole: String
        get() = COLOR_RED

    override val symbol: String
        get() = heartSymbol

}

/**
 * Abstract class representing a Card, a card has a value, and a state (faceUp or faceDown).
 * The type of the card (giving the colorConsole and symbol) properties are implemented by subclasses (like `HeartCard`)
 */
abstract class Card(var isFaceUp: Boolean, var cardValue: CardValue) {
    /**
     * Get the string representation of the card.
     * Example: 10♥ if the card is face up or ** if the card is face down
     * The complicated expression is to add color and some spacing to align 10♥ with 2♥ (make them the same length)
     */
    private val stringRepresentation: String
        get() = if (isFaceUp) "$colorConsole${if(cardValue != CardValue.TEN) " " else ""}${cardValue}$symbol$COLOR_RESET" else " **"

    /**
     * Abstract property representing the color the card will be printed in the console (constants defined in top of screen)
     */
    abstract val colorConsole: String

    /**
     * Abstract property string, the symbol of the card (heart, diamond, club, spade)
     */
    abstract val symbol: String

    /**
     * Flips the card
     */
    fun flip() {
        isFaceUp = !isFaceUp
    }

    /**
     * Those are just to string, equals and hashCode method, it isn't really worth documenting
     */
    override fun toString() = stringRepresentation
    override fun equals(other: Any?): Boolean {
        if (this === other) return true
        if (other !is Card) return false
        if (other.javaClass != this.javaClass)
            return false

        if (isFaceUp != other.isFaceUp) return false
        if (cardValue != other.cardValue) return false

        return true
    }

    override fun hashCode(): Int {
        var result = isFaceUp.hashCode()
        result = 31 * result + cardValue.hashCode()
        return result
    }
}

/**
 * Enumeration corresponding to card values from to two Ace
 */
enum class CardValue(val value: String) {
    ACE("A") {
        override fun next() = TWO
        override fun previous(): CardValue? = null
    },
    TWO("2") {
        override fun next() = THREE
        override fun previous() = ACE
    },
    THREE("3") {
        override fun next() = FOUR
        override fun previous() = TWO
    },
    FOUR("4") {
        override fun next() = FIVE
        override fun previous() = THREE
    },
    FIVE("5") {
        override fun next() = SIX
        override fun previous() = FOUR
    },
    SIX("6") {
        override fun next() = SEVEN
        override fun previous() = FIVE
    },
    SEVEN("7") {
        override fun next() = EIGHT
        override fun previous() = SIX
    },
    EIGHT("8") {
        override fun next() = NINE
        override fun previous() = SEVEN
    },
    NINE("9") {
        override fun next() = TEN
        override fun previous() = EIGHT
    },
    TEN("10") {
        override fun next() = JACK
        override fun previous() = NINE
    },
    JACK("J") {
        override fun next() = QUEEN
        override fun previous() = TEN
    },
    QUEEN("Q") {
        override fun next() = KING
        override fun previous() = JACK
    },
    KING("K") {
        override fun next(): CardValue? = null
        override fun previous() = QUEEN
    };

    companion object {
        fun valueOf(value: Int) = fromString(value.toString())
        fun fromString(value: String) = when(value) {
            "2" -> TWO
            "3" -> THREE
            "4" -> FOUR
            "5" -> FIVE
            "6" -> SIX
            "7" -> SEVEN
            "8" -> EIGHT
            "9" -> NINE
            "10" -> TEN
            "J" -> JACK
            "Q" -> QUEEN
            "K" -> KING
            "A" -> ACE
            else -> valueOf(value)
        }
    }

    override fun toString() = value
    abstract fun next(): CardValue?
    abstract fun previous(): CardValue?

}

class DiamondCard(isFaceUp: Boolean, cardValue: CardValue) : Card(isFaceUp, cardValue) {

    constructor(isFaceUp: Boolean, value: Int) : this(isFaceUp, CardValue.valueOf(value))
    constructor(isFaceUp: Boolean, value: String): this(isFaceUp, CardValue.fromString(value))

    override val colorConsole: String
        get() = COLOR_RED

    override val symbol: String
        get() = "♦"
}

class SpadeCard(isFaceUp: Boolean, cardValue: CardValue) : Card(isFaceUp, cardValue) {

    constructor(isFaceUp: Boolean, value: Int) : this(isFaceUp, CardValue.valueOf(value))
    constructor(isFaceUp: Boolean, value: String): this(isFaceUp, CardValue.fromString(value))

    override val colorConsole: String
        get() = COLOR_BLUE

    override val symbol: String
        get() = "♠"
}

class ClubCard(isFaceUp: Boolean, cardValue: CardValue) : Card(isFaceUp, cardValue) {

    override val colorConsole: String
        get() = COLOR_BLUE

    constructor(isFaceUp: Boolean, value: Int) : this(isFaceUp, CardValue.valueOf(value))
    constructor(isFaceUp: Boolean, value: String): this(isFaceUp, CardValue.fromString(value))

    override val symbol: String
        get() = "♣"
}