package model

import java.util.*
import kotlin.math.min


/**
 * Extension method
 * Removes the n last elements of the list (or all the elements if there are less than n)
 */
fun <T> MutableList<T>.removeNLasts(n: Int): MutableList<T> {
    val res = mutableListOf<T>()
    for(i in min(n, size) downTo 1) {
        res.add(0, get(size - 1))
        removeLast()
    }
    return res
}

/**
 * Class representing a deck, it contains only a list of cards and several methods to manipulate that list
 */
open class Deck(val cards: MutableList<Card> = mutableListOf()) {

    val top: Card?
        get() = if (cards.isNotEmpty()) cards.last() else null

    companion object {
        /**
         * Static method used to create a shuffled Deck of 52 cards.
         * @param faceUp Optional param used to determine if the cards in the deck are visible (faceUp) or hidden (!faceUp)
         */
        @JvmOverloads
        fun deckOf52(faceUp: Boolean = false) = Deck(
            CardValue.values().flatMap { value ->
                setOf(HeartCard::class.java, DiamondCard::class.java, ClubCard::class.java, SpadeCard::class.java).map { cardClass ->
                    cardClass.getConstructor(Boolean::class.java, CardValue::class.java).newInstance(faceUp, value)
                }
            }.shuffled().toMutableList()
        )
    }

    /**
     * Adds a card under the deck
     */
    fun addBelow(card: Card) {
        cards.add(0, card)
    }

    /**
     * Draws the top card from the deck.
     * This method returns the top card of the deck and also removes it from the deck
     */
    open fun draw() = top.also {
        if (cards.isNotEmpty())
            cards.removeLast()
    }

    /**
     * Operator +=
     * Adds a card on top of the deck if the card can go on it
     */
    operator fun plusAssign(card: Card) {
        if(pileCompatibilityCheck(card))
            cards+=card
    }

    /**
     * Operator +=
     * Adds a list of cards on top of the deck.
     * Iterates through all of the cards of the list and adds them one by one, stops if a card is misplaced
     */
    operator fun plusAssign(cards: List<Card>) {
        cards.forEach {
            if(pileCompatibilityCheck(it))
                this.cards+=it
            else
                return
        }
    }

    /**
     * Operator +
     * Convenience function doing the same as `addSingleCardAndReturnNewDeck`
     * TODO: Operator overload
     */
    operator fun plus(card: Card) = addSingleCardAndReturnNewDeck(card)

    /**
     * Creates a new Deck of cards being equal to this deck with the given card on top
     * TODO: If expression
     */
    fun addSingleCardAndReturnNewDeck(card: Card) = if(pileCompatibilityCheck(card)) {
        Deck(cards.apply { add(card) })
    }
    else {
        null
    }

    /**
     * Operator +
     * Creates a new Deck of cards being equal to this deck with the given cards on top
     */
    operator fun plus(cards: List<Card>): Deck? {
        val newDeck = this.javaClass.getConstructor(List::class.java).newInstance(this.cards)
        newDeck += cards
        return newDeck
    }

    /**
     * Shuffles the deck
     */
    fun shuffle() {
        this.cards.shuffle()
    }

    /**
     * Returns the deck as a string, which is:
     * The top card if there is one or __ if no card is visible
     */
    override fun toString() : String {
        val string = top?.toString()
        return when {
            string == null -> " __"
            else -> string
        }
    }

    /**
     * Removes the number last cards of the deck and returns them
     */
    open fun remove(number: Int) = cards.removeNLasts(number)

    /**
     * Removes all the cards from the deck
     */
    fun removeAll() = remove(cards.size)

    /**
     * Method to check that the given card can go on the deck
     * @param card
     */
    open fun pileCompatibilityCheck(card: Card) = true

    /**
     * Flips all the N last cards (preserves cards order)
     */
    fun flipCards(count: Int) {
        for(i in cards.size-count until cards.size )
            cards[i].flip()
    }

    /**
     * Flips the whole deck, flipping each cards and reversing the cards order
     */
    fun flipDeck() {
        cards.reverse()
        cards.forEach { it.flip() }
    }

    /**
     * Flips all the cards in the deck (preserves cards order)
     */
    fun flipCards() {
        flipCards(cards.size)
    }
}

/**
 * SubClass of `Deck` corresponding to the piles of the Tableau (the seven fanned piles where the player can put incoming cards)
 */
class GamePile(cards: MutableList<Card> = mutableListOf()): Deck(cards) {
    /**
     * Overrides the default card check (always set to true)
     * A card can go on the pile if it is of a different color than the top one and is exactly one value down
     */
    override fun pileCompatibilityCheck(card: Card): Boolean {
        // If the pile is empty, then only a king can be put on it
        if(cards.isEmpty())
            return card.cardValue == CardValue.KING

        val validColor = when (cards.last()) {
            is HeartCard, is DiamondCard -> card is SpadeCard || card is ClubCard // Red card on top of pile, new card must be black
            else -> card is HeartCard || card is DiamondCard // Black card on top of pile, new card must be red
        }

        if(!validColor)
            return false

        // The next value of the incoming card is the value of the top card
        return card.cardValue.next()?.equals(top?.cardValue) ?: false
    }

    /**
     * Drawing a card from the pile returns it, removes it from the pile and flips the top card if it is face down
     */
    override fun draw() =  top.also {
        if (cards.isNotEmpty())
            cards.removeLast()
        else
            return@also
        if (top?.isFaceUp == false)
            top!!.flip()
    }

    /**
     * Removes the top N cards from the pile and flips the new top card if it is
     */
    override fun remove(number: Int): MutableList<Card> {
        return super.remove(number).also {
            if(top?.isFaceUp == false)
                top!!.flip()
        }
    }

}

/**
 * SubClass of Deck corresponding to a store deck (the four decks on the top right of the solitaire game)
 */
class StoreDeck<T: Card>: Deck() {

    override fun pileCompatibilityCheck(card: Card): Boolean {
        if(cards.isEmpty())
            return card.cardValue == CardValue.ACE

        val validColor = when (cards.last()) {
            is HeartCard, is DiamondCard -> card is HeartCard || card is DiamondCard
            else -> card is SpadeCard || card is ClubCard
        }

        if(!validColor)
            return false

        return card.cardValue.previous()?.equals(cards.last().cardValue) ?: false
    }

    fun checkCardCanGoOnPile(card: T): Boolean = pileCompatibilityCheck(card)
}