package model

object Game {
    val stack = Deck.deckOf52(false)
    val visibleDeck: Deck = Deck(stack.remove(1)).apply { top!!.flip() }

    val heartPile = StoreDeck<HeartCard>()
    val diamondPile = StoreDeck<DiamondCard>()
    val clubPile = StoreDeck<ClubCard>()
    val spadePile = StoreDeck<SpadeCard>()
    val gamePiles = buildPiles(7)

    override fun toString() =
"""=========================================================================
            $stack  ${visibleDeck}           $heartPile  $diamondPile  $clubPile  $spadePile
      
        
   ${(0 until 7).joinToString("    ")} 
${printPiles()}
=======================================================================
    """.trimIndent()

    private fun printPiles() = StringBuilder().apply {
        val maxIterations = gamePiles.map { pile -> pile.cards.size }.maxOrNull()?:0
        for(i in 0 until maxIterations) {
            for(pile in gamePiles) {
                append("  ")
                append(pile.cards.map { it.toString() }.getOrElse(i) {"   "})
            }
            append(System.lineSeparator())
        }
        toString()
    }

    private fun buildPiles(number: Int): List<Deck> {
        val list = mutableListOf<Deck>()
        for(i in 1 .. number)
            list.add(GamePile(stack.remove(i)).apply { flipCards(1) })
        return list
    }

    fun place(pile: Int) {
        val card = visibleDeck.top?:return
        if(gamePiles[pile].pileCompatibilityCheck(card))
            gamePiles[pile] += visibleDeck.draw()?: return
    }

    fun place(origin: Int, end:Int) {
        val card = gamePiles[origin].top?: return
        if(gamePiles[end].pileCompatibilityCheck(card))
            gamePiles[end] += gamePiles[origin].draw()?: return
    }

    fun move(originPile: Int, rootCardValue: CardValue, targetPileIndex: Int) {
        val cards = gamePiles[originPile].cards
        val targetPile = gamePiles[targetPileIndex]
        val rootCardIndex = cards.filter { it.isFaceUp }.indexOfFirst { it.cardValue == rootCardValue }
        if (rootCardIndex == -1)
            return
        if(targetPile.pileCompatibilityCheck(cards[rootCardIndex])) {
            targetPile += gamePiles[originPile].remove(cards.size - rootCardIndex)
        }
    }

    fun draw() {
        if(stack.top != null)
            visibleDeck.plusAssign(stack.draw()?.apply { flip() }?:return)
        else
            stack += visibleDeck.run {
                flipDeck();
                removeAll()
            }
    }

    fun store(pile: Int) {
        val card = gamePiles[pile].top?:return
        when (card) {
            is HeartCard -> if (heartPile.checkCardCanGoOnPile(card)) heartPile += gamePiles[pile].draw()!!
            is DiamondCard -> if (diamondPile.checkCardCanGoOnPile(card)) diamondPile += gamePiles[pile].draw()!!
            is ClubCard -> if (clubPile.checkCardCanGoOnPile(card)) clubPile += gamePiles[pile].draw()!!
            is SpadeCard -> if (spadePile.checkCardCanGoOnPile(card)) spadePile += gamePiles[pile].draw()!!
        }
    }

    /**
     * Puts the visible deck card on one of the store piles
     * TODO: NullSafety
     * TODO: SmartCast => To show that smartcast works, make them specify explicitely the type of card (card: Card) and try to do `is DiamondCard -> heartPile.checkCardCanGoOnPile(card) and see that it does not work
     */
    fun store() {
        val card =  visibleDeck.top?:return
        when (card) {
            is HeartCard -> if (heartPile.checkCardCanGoOnPile(card)) heartPile += visibleDeck.draw()!!
            is DiamondCard -> if (diamondPile.checkCardCanGoOnPile(card)) diamondPile += visibleDeck.draw()!!
            is ClubCard -> if (clubPile.checkCardCanGoOnPile(card)) clubPile += visibleDeck.draw()!!
            is SpadeCard -> if (spadePile.checkCardCanGoOnPile(card)) spadePile += visibleDeck.draw()!!
        }
    }
}